/*  Copyright (C) 1994 Board of Trustees of the University of Illinois
 *  $Id: perlcci.c,v 1.2 1994/12/14 21:10:32 edburns Exp $
 *  This minimal implementation by Ed Burns <edburns@ncsa.uiuc.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 1, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "EXTERN.h"
#include "perl.h"
#include "cci.h"
#include "port.h"
#include <stdio.h>

/* data to interface with perl */
static enum uservars {
  UV_MCCI_OK,			   /* return codes */
  UV_MCCI_FAIL,
  UV_MCCI_OUTOFMEMORY,
  UV_REQUEST_FAIL,
  UV_NETWORK_ERROR,
  
  UV_MCCI_DEFAULT,		   /* output types, for MCCIGet */
  UV_MCCI_OUTPUT_NONE,
  UV_MCCI_OUTPUT_CURRENT,
  UV_MCCI_OUTPUT_NEW,
  UV_MCCI_ABSOLUTE,
  UV_MCCI_RELATIVE,
};

/* need one of these for each CCI function */
static enum usersubs {
  US_MCCIInitialize,
  US_MCCIConnect,
  US_MCCIDisconnect,
  US_MCCIGetSocketDescriptor,
  US_MCCIIsConnected,
  US_MCCIPoll,
  US_MCCIIsThereInput,
  US_MCCISendAnchor,
  US_MCCIGet,
  US_MCCISendOutput,
  US_MCCISendOutputStop,
};

/* functions to interface with perl */
int userinit();			   /* this is the unresolved name in uperl.o */
int init_perlcci();
static int usersub();
static int userval();
     
/* functions to facilitate perl stack manipulation */
int doArgs();
void str_ptrset();
void *str_gptr();

/* functions to handle callBacks into perl:
 * Note: need one of these for each CCI function
 *       that uses a callBack
 */
void callv();
void anchorCallBack();
void sendOutputCallBack();

int userinit() {
  init_perlcci();
}

int init_perlcci() {
  struct ufuncs uf;
  char *filename = "perlcci.c";
  
  uf.uf_val = userval;
  
#define MAGICVAR(name, ix) uf.uf_index = ix, magicname(name, &uf, sizeof uf)

  MAGICVAR("MCCI_OK",		   UV_MCCI_OK); 
  MAGICVAR("MCCI_FAIL",		   UV_MCCI_FAIL);
  MAGICVAR("MCCI_OUTOFMEMORY",	   UV_MCCI_OUTOFMEMORY);
  MAGICVAR("REQUEST_FAIL",	   UV_REQUEST_FAIL);
  MAGICVAR("MCCI_DEFAULT",	   UV_MCCI_DEFAULT);
  MAGICVAR("MCCI_OUTPUT_NONE",	   UV_MCCI_OUTPUT_NONE);
  MAGICVAR("MCCI_OUTPUT_CURRENT",  UV_MCCI_OUTPUT_CURRENT);
  MAGICVAR("MCCI_OUTPUT_NEW",      UV_MCCI_OUTPUT_NEW);
  MAGICVAR("MCCI_ABSOLUTE",        UV_MCCI_ABSOLUTE);
  MAGICVAR("MCCI_RELATIVE",	   UV_MCCI_RELATIVE);

  make_usub("MCCIInitialize", US_MCCIInitialize, usersub, filename);
  make_usub("MCCIConnect", US_MCCIConnect, usersub, filename);
  make_usub("MCCIDisconnect", US_MCCIDisconnect, usersub, filename);
  make_usub("MCCIGetSocketDescriptor", US_MCCIGetSocketDescriptor, usersub, filename);
  make_usub("MCCIIsConnected", US_MCCIIsConnected, usersub, filename);
  make_usub("MCCIPoll", US_MCCIPoll, usersub, filename);
  make_usub("MCCIIsThereInput", US_MCCIIsThereInput, usersub, filename);
  make_usub("MCCISendAnchor", US_MCCISendAnchor, usersub, filename);
  make_usub("MCCIGet", US_MCCIGet, usersub, filename);
  make_usub("MCCISendOutput", US_MCCISendOutput, usersub, filename);
  make_usub("MCCISendOutputStop", US_MCCISendOutputStop, usersub, filename);
}

static int usersub(ix, sp, items)
int ix;
register int sp;
register int items;
{
  STR **st = stack->ary_array + sp;
  register int i;
  register char *tmps;
  register STR *Str;		   /* used in str_get and str_gnum macros */
  
  switch (ix) {
  case US_MCCIInitialize:
    if (items !=0)
      fatal("Usage: &MCCIInitialize");
    else {
      int retval;
#ifdef DEBUG
      debugger();
#endif
      retval = MCCIInitialize();
      str_numset(st[0], (double) retval);
    }
    return sp;

  case US_MCCIConnect:
    if (items < 4)
      fatal("Usage: &MCCIConnect($serverAddress, $port, $callBack, @callBackData");
    else {
      MCCIPort perlPort;
      char*  addr = (char*) str_get(st[1]);
      int    port = (int) str_gnum(st[2]);
      void*  perlCallBackData;
      if (MCCI_OK != doArgs((char **) &perlCallBackData, items - 2, sp))
	fatal("CCI: unable to allocate storage for callback data");
      perlPort = (MCCIPort) MCCIConnect(addr, port, callv, 
				      (void *) perlCallBackData);
      str_ptrset(st[0], perlPort);
    }
    return sp;

  case US_MCCIDisconnect:
    if (items != 1)
      fatal("Usage: &MCCIDisconnect($serverPort)");
    else {
      int retval;
      MCCIPort perlPort = (MCCIPort) str_gptr(st[1]); 

      retval = MCCIDisconnect(perlPort);
      str_numset(st[0], (double) retval);
    }
    return sp;

  case US_MCCIGetSocketDescriptor:
    if (items != 1)
      fatal("Usage: &MCCIGetSocketDescriptor($serverPort)");
    else {
      int retval;
      MCCIPort usrPort = (MCCIPort) str_gptr(st[1]); 

      retval = MCCIGetSocketDescriptor(usrPort);
      str_numset(st[0], (double)retval);
    }  
    return sp;

  case US_MCCIIsConnected:
    if (items != 1)
      fatal("Usage: &MCCIIsConnected($serverPort)");
    else {
      int retval;
      MCCIPort usrPort = (MCCIPort) str_gptr(st[1]);

      retval = MCCIIsConnected(usrPort);
      str_numset(st[0], (double) retval);
    }  
    return sp;

  case US_MCCIPoll:
    if (items != 1)
      fatal("Usage: &MCCIPoll($serverPort)");
    else {
      int retval;
      MCCIPort usrPort = (MCCIPort) str_gptr(st[1]);

      retval = MCCIPoll(usrPort);
      str_numset(st[0], (double) retval);
    }  
    return sp;
    
  case US_MCCIIsThereInput:
    if (items != 1)
      fatal("Usage: &MCCIIsThereInput($serverPort)");
    else {
      int retval;
      MCCIPort usrPort = (MCCIPort) str_gptr(st[1]);

      retval = MCCIIsThereInput(usrPort);
      str_numset(st[0], (double) retval);
    }  
    return sp;
    
  case US_MCCISendAnchor:
    if (items < 4)
      fatal("Usage: &MCCISendAnchor($serverPort, $on, $callBack, @callBackData)");
    else {
      int retval;
      MCCIPort usrPort = (MCCIPort) str_gptr(st[1]);
      int on = str_gnum(st[2]);
      void*  perlCallBackData;     

      if (MCCI_OK != doArgs((char **) &perlCallBackData, items - 2, sp))
	fatal("CCI: unable to allocate storage for callback data");
      retval = MCCISendAnchor(usrPort, on, anchorCallBack, 
						(void *) perlCallBackData);
      str_numset(st[0], (double) retval);
    }  
    return sp;
 
  case US_MCCIGet:
    if (items < 4)
      fatal("Usage: &MCCIGet($serverPort, $uri, $output, $absRelative, $additionalHeader)");
    else {
      int retval;
      MCCIPort usrPort = (MCCIPort) str_gptr(st[1]);
      char *uri = str_get(st[2]);
      int output = str_gnum(st[3]);
      int absRelative = str_gnum(st[4]);
      char *additionalHeader = str_get(st[5]);
#ifdef HACK
      additionalHeader = NULL;
#endif
      retval = MCCIGet(usrPort, uri, output, absRelative, additionalHeader);
      str_numset(st[0], (double) retval);
    }  
    return sp;

  case US_MCCISendOutput:
    if (items < 4)
       fatal("Usage: &MCCISendOutput($serverPort, $mimeType, $callback, @callBackData)");
    else {
      int retval;
      MCCIPort usrPort = (MCCIPort) str_gptr(st[1]);
      char *mimeType = str_get(st[2]);
      void *sendOutputCallBackData;

      if (MCCI_OK != doArgs((char **) &sendOutputCallBackData, items - 2, sp))
	fatal("CCI: unable to allocate storage for callback data");
      retval = MCCISendOutput(usrPort, mimeType, sendOutputCallBack,
			      (void *) sendOutputCallBackData);
      str_numset(st[0], (double) retval);
    }
    return sp;
  
  case US_MCCISendOutputStop:
    if (items != 2)
      fatal("Usage: $MCCISendOutputStop($serverPort, $mimeType)");
    else {
      int retval;
      MCCIPort usrPort = (MCCIPort) str_gptr(st[1]);
      char* mimeType = str_get(st[2]);

      retval = MCCISendOutputStop(usrPort, mimeType);
      str_numset(st[0], (double) retval);
    }
    return sp;

  default:
   fatal("Unimplemented user-defined subroutine");
  }
    return sp;
}
 
static int
userval(ix, str)
int ix;
STR *str;
{
    switch (ix) {
    case UV_MCCI_OK:			   /* return codes */
      str_numset(str, (double)MCCI_OK);
      break;
    case UV_MCCI_FAIL:
      str_numset(str, (double)MCCI_FAIL);
      break;
    case UV_MCCI_OUTOFMEMORY:
      str_numset(str, (double)MCCI_OUTOFMEMORY);
      break;
    case UV_REQUEST_FAIL:
      str_numset(str, (double)MCCI_REQUEST_FAIL);
      break;
    case UV_NETWORK_ERROR:
      str_numset(str, (double)MCCI_NETWORK_ERROR);
      break;
    case UV_MCCI_DEFAULT:
      str_numset(str, (double)MCCI_DEFAULT);
      break;
    case UV_MCCI_OUTPUT_NONE:
      str_numset(str, (double)MCCI_OUTPUT_NONE);
      break;
    case UV_MCCI_OUTPUT_CURRENT:
      str_numset(str, (double)MCCI_OUTPUT_CURRENT);
      break;
    case UV_MCCI_OUTPUT_NEW:
      str_numset(str, (double)MCCI_OUTPUT_NEW);
      break;
    case UV_MCCI_ABSOLUTE:
      str_numset(str, (double)MCCI_ABSOLUTE);
      break;
    case UV_MCCI_RELATIVE:
      str_numset(str, (double)MCCI_RELATIVE);
      break;
    }
    return 0;
  }

int doArgs(arglist, argc, sp)
  char **arglist;
  int  argc;                                  /* perlSub + numargs(perlSub) */
  int  sp;
{
/* pre-condition: perl stack has values from st[3] to st[3+(argc-1)]
 *                st[3] is the name of the perl subroutine
 *                st[4] thru st[3+(argc-1)] are the args to st[3]
 *
 * post-condition: argv[] is an array of pointers to characters suitable
 *                 to be pointed to by the callBackData element of an MCCIPort
 *                 structure.  
 *                 Memory is allocated for the array and each of the elements
 *                 in it.  This memory should be freed.
 */
 
  STR **st = stack->ary_array + sp;
  char  **argv, spPtr[4], gimmePtr[4];        /* tmp vars for perl stuff */
  int i;

  if (!(*arglist = calloc((argc+3),
			  sizeof(char *)))) { 
    return(MCCI_OUTOFMEMORY);
  }
  sprintf(spPtr, "%d\0", sp);               /* perl stack pointer */
  sprintf(gimmePtr, "1\0");                 /* scalar context */

  argv = (char **) *arglist;	
  if (!(argv[0] = calloc(strlen((char*) str_get(st[3])) + 1, 
			 sizeof(char)))) {
    return(MCCI_OUTOFMEMORY);
  }
  strcpy(argv[0], (char*) str_get(st[3]));

  if (!(argv[1] = calloc(strlen(spPtr) + 1, 
			 sizeof(char)))) { /**** malloc(strlen(spPtr) + 5); ***/
    return(MCCI_OUTOFMEMORY);
  }
  strcpy(argv[1], spPtr); 
  
  if (!(argv[2] = calloc(strlen(gimmePtr) + 1, 
			 sizeof(char)))) {
    return(MCCI_OUTOFMEMORY);
  }
  strcpy(argv[2], gimmePtr);
  
  for (i = 3; i < argc + 2; i++) {          /* copy perlSubArgs */
    if (!(argv[i] = calloc(strlen((char *) 
				  str_get(st[i+1])) + 1, sizeof(char)))) {
      return(MCCI_OUTOFMEMORY);
    }
    strcpy(argv[i], (char *) str_get(st[i+1]));
  }
  argv[i] = NULL;
  
  return(MCCI_OK);
}

void str_ptrset(str, ptr)
  register STR *str;
  void *ptr;
{
  char *buf = malloc(sizeof(char *) + 1);

  if (buf ) {
    sprintf(buf, "0x%lx", ptr);	/* was "%p" */
    str_set(str, buf);
  }
}

void *str_gptr(str)
  register STR *str;
{
  void *ptr;
  char *buf = str_get(str); 
  
  if (buf) { 
    sscanf(buf, "0x%lx", &ptr);	/* was "%p" */
    return(ptr);
  }
}

int callback(subname, sp, gimme, hasargs, numargs)
     char *subname;
     int sp;			   /* stack pointer after args are pushed */
     int gimme;			   /* called in array or scalar context */
     int hasargs;		   /* whether to create a @_ array for routine */
     int numargs;		   /* how many args are pushed on the stack */
{
  static ARG myarg[3];	   /* fake syntax tree node */
  int arglast[3];
  
  arglast[2] = sp;
  sp -= numargs;
  arglast[1] = sp--;
  arglast[0] = sp;
  
  if (!myarg[0].arg_ptr.arg_str)
    myarg[0].arg_ptr.arg_str = str_make("",0);

  myarg[1].arg_type = A_WORD;
  myarg[1].arg_ptr.arg_stab = stabent(subname, FALSE);
  
  myarg[2].arg_type = hasargs ? A_EXPR : A_NULL;
  
  return do_subr(myarg, gimme, arglast);
}

void callv(port, callBackData)
     MCCIPort port;
     void *callBackData[];	   /* null terminated arg list, NULL for none */
{
  char *subname, **argv;
  register int sp, gimme, items = 0;
  int hasargs = 1;
  
  subname = callBackData[0];
  sp = atoi(callBackData[1]); 
  gimme = atoi(callBackData[2]);
  argv = (char **) &callBackData[3];
  
  astore(stack, ++sp, Nullstr);	   /* reserve spot for 1st return arg */
  if (hasargs) {
    while (*argv) {
      astore(stack, ++sp, str_2mortal(str_make(*argv,0)));
      items++;
      argv++;
    }
  }
  callback(subname, sp, gimme, hasargs, items);
}

void anchorCallBack(anchor, callBackData)
     char *anchor;
     void *callBackData[];	   /* null terminated arg list, NULL for none */
{
  char *subname, **argv;
  register int sp, gimme, items = 0;
  int hasargs = 1;
  
  subname = callBackData[0];
  sp = atoi(callBackData[1]);
  gimme = atoi(callBackData[2]);
  argv = (char **) &callBackData[3];
  
  astore(stack, ++sp, Nullstr);	   /* reserve spot for 1st return arg */
  astore(stack, ++sp, str_2mortal(str_make(anchor,0)));
  items++;
  if (hasargs) {
    while (*argv) {
      astore(stack, ++sp, str_2mortal(str_make(*argv,0)));
      items++;
      argv++;
    }
  }
  callback(subname, sp, gimme, hasargs, items);
}

void sendOutputCallBack(mimeType, data, length, callBackData)
     char *mimeType;
     char *data;
     int length;
     void *callBackData[];
{
  char *subname, **argv, buff[80];
  register int sp, gimme, items = 0;
  int hasargs = 1;
  
  subname = callBackData[0];
  sp = atoi(callBackData[1]);
  gimme = atoi(callBackData[2]);
  argv = (char **) &callBackData[3];
  
  astore(stack, ++sp, Nullstr);	   /* reserve spot for 1st return arg */
  
  astore(stack, ++sp, str_2mortal(str_make(mimeType,0)));
  items++;
  astore(stack, ++sp, str_2mortal(str_make(data,length)));
  items++;
  sprintf(buff, "%d", length);
  astore(stack, ++sp, str_2mortal(str_make(buff,0)));
  items++;
  
  if (hasargs) {
    while (*argv) {
      astore(stack, ++sp, str_2mortal(str_make(*argv,0)));
      items++;
      argv++;
    }
  }
  callback(subname, sp, gimme, hasargs, items);
}

